%   Module : ctypes
%   Author : Richard A. O'Keefe
%   Updated: 2/10/87
%   Purpose: Character classification

/*  The predicates to_lower, to_upper, is_alnum, is_alpha, is_cntrl,
    is_digit, is_graph, is_lower, is_upper, is_print, is_punct, and
    is_space are taken directly from the April 84 draft of the C
    standard.  The remaining ones are of my own invention, but are
    reasonably useful.  If you want to make your programs portable
    between different operating systems, use is_endline and is_endfile
    instead of the literal constants 31 and 26 or 10 and -1.

    Xerox Quintus Prolog naturally uses the Xerox XNS character set.
    The Prolog tokeniser regards characters outside the ASCII range
    0..127 as lower-case letters.  However, enumerating roughly 2^16
    answers to is_lower/1 doesn't seem like a good idea.  Quintus
    Prolog has to cope with several character sets:
	ASCII
	EBCDIC
	DEC's Multinational Character Set
	DIS 8859/1 (the new ISO standard, almost identical to DEC's set)
	ASCII KANJI (two versions)
	XNS
    There are no real implementation difficulties, but there are fairly
    extreme difficulties of specification.  See the Xerox XNS document
    for some of the questions.  In the mean-time, this package rejects
    characters outside the ASCII range.
*/
:- module(ctypes, [
	is_alnum/1,
	is_alpha/1,
	is_ascii/1,
	is_cntrl/1,
	is_csym/1,
	is_csymf/1,
	is_digit/1,
	is_digit/3,
	is_endfile/1,
	is_endline/1,
	is_graph/1,
	is_lower/1,
	is_newline/1,
	is_newpage/1,
	is_paren/2,
	is_period/1,
	is_print/1,
	is_punct/1,
	is_quote/1,
	is_space/1,
	is_upper/1,
	to_lower/2,
	to_upper/2
   ]).
:- use_module(library(between), [
	between/3
   ]).

sccs_id('"@(#)87/04/12 ctypes.pl	99.1"').


is_alnum(C) :-			% Letter or Digit
	var(C),
	!,
	(   between(97 /* a */, 122 /* z */, C)
	;   between(65 /* A */,  90 /* Z */, C)
	;   between(48 /* 0 */,  57 /* 9 */, C)
	).
is_alnum(C) :-
	(   C >= 97 /* a */ -> C =< 122 /* z */
	;   C >= 65 /* A */ -> C =<  90 /* Z */
	;   C >= 48 /* 0 */,   C =<  57 /* 9 */
	).



is_alpha(C) :-			% Upper or Lower case letter
	var(C),
	!,
	(   between(97 /* a */, 122 /* z */, C)
	;   between(65 /* A */,  90 /* Z */, C)
	).
is_alpha(C) :-
	(   C >= 97 /* a */ -> C =< 122 /* z */
	;   C >= 65 /* A */,   C =<  90 /* Z */
	).



is_ascii(C) :-
	(   var(C) -> between(0, 127, C)
	;   C >= 0, C =< 127
	).



is_cntrl(C) :-
	var(C),
	!,
	(   C = 127			% DEL, or any other
	;   between(  0,  31, C)	% ASCII control character
	).
is_cntrl(C) :-
	(   C < 32 -> C >= 0		% ASCII control character
	;   C =:= 127			% DEL
	).



is_csym(C) :-			% Letter, Digit, or Underscore
	var(C),
	!,
	(   between(97 /* a */, 122 /* z */, C)
	;   C = 95 /* _ */
	;   between(65 /* A */,  90 /* Z */, C)
	;   between(48 /* 0 */,  57 /* 9 */, C)
	).
is_csym(C) :-
	(   C >= 97 /* a */ -> C =< 122 /* z */
	;   C=:= 95 /* _ */ -> true
	;   C >= 65 /* A */ -> C =<  90 /* Z */
	;   C >= 48 /* 0 */,   C =<  57 /* 9 */
	).



is_csymf(C) :-			% Letter or Underscore
	var(C),
	!,
	(   between(97 /* a */, 122 /* z */, C)
	;   C = 95 /* _ */
	;   between(65 /* A */,  90 /* Z */, C)
	).
is_csymf(C) :-
	(   C >= 97 /* a */ -> C =< 122 /* z */
	;   C=:= 95 /* _ */ -> true
	;   C >= 65 /* A */ ,  C =<  90 /* Z */
	).



is_digit(C) :-			% Decimal digit
	between(48 /* 0 */, 57 /* 9 */, C).



is_digit(C, Base, Weight) :-
	between(2, 36, Base),
	(   var(C) ->
		X is Base-1,
		between(0, X, Weight),
		(   Weight < 10 -> C is Weight + 48 /* 0 */
		;   C is Weight + 87	/* `a`-10 */
		;   C is Weight + 55	/* `A`-10 */
		)
	;   C =< 57 /* 9 */ ->
		C >= 48 /* 0 */, Weight is C-48 /* 0 */, Weight < Base
	;   C =< 90 /* Z */ ->
		C >= 65 /* A */, Weight is C-55 /* A'*/, Weight < Base
	;   C =< 122 /* z */,
		C >= 97 /* a */, Weight is C-87 /* a'*/, Weight < Base
	).



is_endfile(-1).			% was 26 in Dec-10 and C Prolog



is_endline(C) :-		% line terminator, NOT just newline
	C < 32,			% covers EOF, ^G, ^J, ^K, ^L, ^M, ^Z, ESC, ^_
	C =\= 9 /* ^I */.	% not ^I though.


is_graph(C) :-			% makes a mark when you print it
	between(33, 126, C).



is_lower(C) :-			% Lower case letter
	between(97 /* a */, 122 /* z */, C).



is_newline(13 /* ^M */).	% Was 31 in Dce-10 Prolog, is 10 on UNIX.



is_newpage(12 /* ^L */).	% may fail (O/S-dependent)



is_paren( 40 /* ( */,   41 /* ) */).
is_paren( 91 /* [ */,   93 /* ] */).
is_paren(123 /* { */,  125 /* } */).



is_period(46 /* . */).		% a period is anything that ends a sentence
is_period(63 /* ? */).		% (also known as a period).  . is a full stop.
is_period(33 /* ! */).



is_print(C) :-
	between(32, 126, C).



is_punct(C) :-
	var(C),
	!,
	(   between( 32,  47, C)	% between space and digits
	;   between( 58,  64, C)	% between digits and uppers
	;   between( 91,  96, C)	% between uppers and lowers
	;   between(123, 126, C)	% between lowers and delete
	).
is_punct(C) :-
	(   C < 48 /* 0 */ -> C >=32 /*   */
	;   C < 65 /* A */ -> C > 57 /* 9 */
	;   C < 97 /* a */ -> C > 90 /* Z */
	;   C >122 /* z */,   C <127 /* ^? */
	).



is_quote(39 /* ' */).
is_quote(34 /* " */).
is_quote(96 /* ` */).



is_space(32).			% ` `
is_space(10).			% -maybe- `\n`
is_space( 9).			% `\t`
is_space(31).			% `\n` in Dec-10 Prolog
is_space(11).			% `\v`
is_space(12).			% `\f`
is_space(13).			% `\r`



is_upper(C) :-
	between(65 /* A */, 90 /* Z */, C).



is_white(32).			% space
is_white( 9).			% tab



%   to_lower(UpperCase, LowerCase) and
%   to_upper(LowerCase, UpperCase)
%   convert letters from one case to another.  Other characters
%   they are supposed to leave alone.

to_lower(U, L) :-
	between(0, 127, U),
	(   U >= 65 /* A */, U =< 90 /* Z */ -> L is U+32
	;   /* otherwise */ L = U
	).


to_upper(L, U) :-			% CHANGED IN 1.6 FINAL RELEASE!
	between(0, 127, L),
	(   L >= 97 /* a */, L =< 122 /* z */ -> U is L-32
	;   /* otherwise */ U = L
	).

